/*
   Class Name:

      CPlane.

   Created by:

      Allen Sherrod (Programming Ace of www.UltimateGameProgramming.com).

   Description:

      This class represents a plane in space.
*/


#include"CPlane.h"


CPlane::CPlane()
{
   // Initialize variables.
   a = 0; b = 0; c = 0; d = 0;
}


CPlane::CPlane(float A, float B, float C, float D)
{
   // Initialize variables.
   a = A; b = B; c = C; d = D;
}


CPlane::CPlane(CVector4 n, float D)
{
   // Initialize variables.
   a = n.x; b = n.y; c = n.z; d = D;
}


CPlane::CPlane(const CPlane &p)
{
   // Initialize variables.
   a = p.a; b = p.b; c = p.c; d = p.d;
}


void CPlane::CreatePlaneFromTri(CVector4 tri1, CVector4 tri2, CVector4 tri3)
{
   // This function will create a plane out of a triangle.  First we normalize
   // the triangle.
   CVector4 normal;
   normal.CrossProduct(tri3 - tri1, tri2 - tri1);
   normal.Normal();

   // Save the normal data to our plane data.  To calculate the d we take the
   // dot product of the triangle normal with the first triangle.
   a = normal.x;
   b = normal.y;
   c = normal.z;
   d = - (a * tri1.x + b * tri1.y + c * tri1.z);
}


void CPlane::SetPlaneNormal(float A, float B, float C)
{
   // Simply set the data members.
   a = A; b = B; c = C;
}


void CPlane::SetPlaneIntercept(float D)
{
   // Set only d.
   d = D;
}


bool CPlane::Intersect(CPlane p2, CPlane p3, CVector4 &intersectPoint)
{
   // This will find the interection of 3 planes and store it in
   // the vector intersectPoint.

   // First we are just placing all data in temp CVector4 objects.
   CVector4 temp;
   CVector4 normal(a, b, c, d);
   CVector4 p2Normal(p2.a, p2.b, p2.c, p2.d);
   CVector4 p3Normal(p3.a, p3.b, p3.c, p3.d);

   // Get cross product of p2 and p3.
   temp.CrossProduct(p2Normal, p3Normal);

   // Get the dot product of this plane with the cross product of p2 and p3.
   float dp = normal.DotProduct3(temp);

   // If dp is 0 then it is safe to return right away.
   if(dp == 0.0f) return false;

   // Else we must find intersect point.
   temp.CrossProduct(p2Normal, p3Normal);
   intersectPoint = temp * d;

   temp.CrossProduct(p3Normal, normal);
   intersectPoint = intersectPoint + (temp * p2.d);

   temp.CrossProduct(normal, p2Normal);
   intersectPoint = intersectPoint + (temp * p3.d);

   intersectPoint = intersectPoint / -dp;

   return true;
}


int CPlane::ClassifyPoint(CVector4 &v)
{
   // This function will determine if a point is on, in front of, or behind
   // the plane.  First we store the dot product of the plane and the point.
   float distance = a * v.x + b * v.y + c * v.z + d;

   // Simply put if the dot product is greater than 0 then it is infront of it.
   // If it is less than 0 then it is behind it.  And if it is 0 then it is on it.
   if(distance > 0.001) return FRONT_PLANE;
   if(distance < -0.001) return BEHIND_PLANE;

   return ON_PLANE;
}


int CPlane::ClassifyPoint(float x, float y, float z)
{
   // This function will determine if a point is on, in front of, or behind
   // the plane.  First we store the dot product of the plane and the point.
   float distance = a * x + b * y + c * z + d;

   // Simply put if the dot product is greater than 0 then it is infront of it.
   // If it is less than 0 then it is behind it.  And if it is 0 then it is on it.
   if(distance > 0.001) return FRONT_PLANE;
   if(distance < -0.001) return BEHIND_PLANE;

   return ON_PLANE;
}


void CPlane::Lerp(CPlane p1, CPlane p2, float amount)
{
   // This function will perform a linear interpolation between two planes
   // and store the result in this plane.

   // First calculate a, b, c, d.
   a = p1.a * (1.0f - amount) + p2.a * amount;
   b = p1.b * (1.0f - amount) + p2.b * amount;
   c = p1.c * (1.0f - amount) + p2.c * amount;
   d = p1.d * (1.0f - amount) + p2.d * amount;

   // Next we will normalize the plane.  First we get the squre root of the length
   // then divide each componet (a, b, c) from it to normalize.
   float length = (float)sqrt((a * a + b * b + c * c));
   
   if(length == 0.0f) length = 1.0f;
   length = 1 / length;

   // Normalize all but d.
   a = a * length;
   b = b * length;
   c = c * length;
}


float CPlane::GetDistance(CVector4 &v)
{
   // This function will simply perform a dot product of the point and
   // this plane.
   return a * v.x + b * v.y + c * v.z + d;
}


float CPlane::GetDistance(float x, float y, float z)
{
   // This function will simply perform a dot product of the point and
   // this plane.
   return a * x + b * y + c * z + d;
}


// Copyright December 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com